/****************************************************************************
**  Version: 20110908
**
**  This file contains the SAS macros that aid in moving EG/AMO/STP
**  code to the grid.
**
**  gridClientPrecode & gridOneTimeSetup can take the following optional 
**  parameters:
**
**    server=<sas_app_server_name>  
**
**       gridClientcode will try to determine the name of the application 
**       server from the _SASSERVERNAME macro variable or, if that does not 
**       exist, it will use the values of the METAAUTORESOURCES option.
**       If neither are acceptable, set this parameter to the SAS application
**       server name that has the logical grid server to use.
**                                 
**    jobname=<name_of_grid_job>    
**    
**       gridClientcode will create the name of the grid job from the
**       application using the code (if it can be determined), the type
**       of code (STP or other), the user and the SAS process ID. If none
**       of these are acceptable, this parameter can be used to specify
**       the name of the grid job to use.
**
**    debug=<y|Y|1>
**
**       Debugging information can be output by setting this option to
**       either "y" or "Y" or "1" (without the quotes).
**
**
**  gridClientPostcode can take the following optional parameter:
**
**    signoff=<y|Y>
**
**       By default, the client post code does not signoff the grid. Setting
**       this parameter to "y" or "Y" (without the quotes) will force the 
**       client to SIGNOFF
**
****************************************************************************/
/*-------------------------------------------------------------------------
** Macro to start debug output based on _GRIDDEBUG
*/
%macro gridDebugBegin;

  %global _GRIDDEBUG;

  %global _GRIDNOTESOPT;
  %global _GRIDSOURCEOPT;
  %global _GRIDSOURCE2OPT;
  %global _GRIDMPRINTOPT;

  %let _GRIDNOTESOPT   = %sysfunc(getoption(notes));
  %let _GRIDSOURCEOPT  = %sysfunc(getoption(source));
  %let _GRIDSOURCE2OPT = %sysfunc(getoption(source2));
  %let _GRIDMPRINTOPT  = %sysfunc(getoption(mprint));

  %if (&_GRIDDEBUG) %then
    %do;
      OPTIONS SOURCE;
      OPTIONS SOURCE2;
      OPTIONS NOTES;
      OPTIONS MPRINT;
    %end;
  %else
    %do;
      OPTIONS NOSOURCE;
      OPTIONS NOSOURCE2;
      OPTIONS NONOTES;
      OPTIONS NOMPRINT;
    %end;

%mend;

/*-------------------------------------------------------------------------
** Macro to end debug output started by gridDebugBegin
*/
%macro gridDebugEnd;

  %global _GRIDNOTESOPT;
  %global _GRIDSOURCEOPT;
  %global _GRIDSOURCE2OPT;
  %global _GRIDMPRINTOPT;

  options &_GRIDMPRINTOPT;
  options &_GRIDNOTESOPT;
  options &_GRIDSOURCE2OPT;
  options &_GRIDSOURCEOPT;

%mend;

/*-------------------------------------------------------------------------
** Macro on client to do one-time setup for the grid by 
**   1) Executing a signon to establish a grid session
**   2) Do all the required processing to link the client and the 
**      remote CONNECT server in the grid.
**   3) Instantiate client macros used every time 
*/
%macro gridOneTimeSetup(server=,jobname=,debug=n);

  %global _GRIDSETUPCOMPLETED;
  %global _SASSERVERNAME;
  %global _CLIENTNAME;
  %global _GRIDDEBUG;
  %global _GRIDJOBNAME;
  %global _GRIDSIGNONRC;

  %if ("&_GRIDSETUPCOMPLETED" eq "") %then
    %do;
      %let _GRIDSETUPCOMPLETED=1;

      /*-------------------------------------------------------------------
      ** Setup the debug macro 
      */
      %if ("&debug" eq "y") or ("&debug" eq "Y") or ("&debug" eq "1") %then
        %do;
          %let _GRIDDEBUG=1;
          %put GRID: Grid debugging enabled;
        %end;
      %else
        %let _GRIDDEBUG=0;

      %gridDebugBegin;

      /*-------------------------------------------------------------------
      ** If _SASSERVERNAME is not specified, try looking at the 
      ** METAAUTORESOURCES option for its value.
      */
      %if ("&server" ne "") %then
        %let _SASSERVERNAME=&server;
      %else %if ("&_SASSERVERNAME" eq "") %then
        %let _SASSERVERNAME=%sysfunc(getoption(metaautoresources));

      /*-------------------------------------------------------------------
      ** Create the job name from the client, user, and sas process ID
      */
      %if ("&jobname" ne "") %then
        %let _GRIDJOBNAME=&jobname;
      %else %if ("&_CLIENTNAME" eq "") %then
        %do;
          %if ("&_CLIENTAPP" eq "") %then
            %let _GRIDJOBNAME=SASGrid.&sysuserid..&sysjobid;
          %else %if ("&_CLIENTAPP" eq "'SAS Enterprise Guide'") %then
            %let _GRIDJOBNAME=EG.&sysuserid..&sysjobid;
          %else %if ("&_CLIENTAPP" eq "'SAS Add-In for Microsoft Office'") %then
            %let _GRIDJOBNAME=AMO.&sysuserid..&sysjobid;
          %else
            %let _GRIDJOBNAME=SASGrid.&sysuserid..&sysjobid;
        %end;
      %else %if ("&_CLIENTNAME" eq "SAS Enterprise Guide") %then
        %let _GRIDJOBNAME=EG.STP.&sysuserid..&sysjobid;
      %else %if ("&_CLIENTNAME" eq "SAS Add-In for Microsoft Office") %then
        %let _GRIDJOBNAME=AMO.STP.&sysuserid..&sysjobid;
      %else
        %let _GRIDJOBNAME=SASGrid.&sysuserid..&sysjobid;

      /*-------------------------------------------------------------------
      ** Signon to the grid
      */
      %if ("&SYSVER" eq "9.1") %then
        %let rc=%sysfunc(grdsvc_enable(grid,"resource=&_SASSERVERNAME"));
      %else
        %let rc=%sysfunc(grdsvc_enable(grid,"resource=&_SASSERVERNAME;jobname=_GRIDJOBNAME;")); 

      %put ;
      %put ;

      /*-------------------------------------------------------------------
      ** Output the results
      */
      %if ("&rc" ne "0") %then
        %do;
          %put ;
          %put ;
          %put *********************************************************************;
          %put GRID: Grid Enable to Application Server "&_SASSERVERNAME", RC=&rc;
          %put GRID: ERROR - Grid not enabled. Exitting.;
          %put *********************************************************************;
          %put ;
          %let _GRIDSETUPCOMPLETED=;
          %return;
        %end;
      %else
        %do;
          %put ;
          %put ;
          %put *********************************************************************;
          %put GRID: Grid Enable to Application Server "&_SASSERVERNAME", RC=&rc;
          %put GRID: Grid job will be named "&_GRIDJOBNAME";
          %put GRID: Workspace server WORK is %sysfunc(pathname(WORK));
          %put *********************************************************************;
          %put ;
        %end;

      /*-------------------------------------------------------------------
      ** Try to SIGNON
      */
      %let _GRIDSIGNONRC=0;
      signon grid cmacvar=_GRIDSIGNONRC;

      /*-------------------------------------------------------------------
      ** If the SIGNON failed, reset everything for the next try
      */
      %if ("&_GRIDSIGNONRC" ne "0") %then
        %do;
          %put GRID: ERROR - SIGNON failed. Exitting.;
          %let _GRIDSETUPCOMPLETED=;
          %return;
        %end;

      /*-------------------------------------------------------------------
      ** Send the debug flag to the remote server
      */
      %syslput _GRIDDEBUG=&_GRIDDEBUG /remote=grid;

      /*-------------------------------------------------------------------
      ** Remap the grid server's work directory to RMTWORK so the user 
      ** can see stuff in the grid work directory. 
      */
      libname rmtwork slibref=work server=grid;

      /*-------------------------------------------------------------------
      ** Copy all the macros in our work directory to the server's work
      ** directory. We move the macro libraries at the proper level. For
      ** a workspace server this is usually WORK.SASMAC1. We cannot move
      ** it directly to WORK.SASMACR since it has a lock on it and we
      ** will get an error so we move it to RMTWORK.SASMACC
      */
      %if (%sysfunc(exist(WORK.SASMACR,CATALOG))) %then
        %do;
          %if (&_GRIDDEBUG) %then
            %put Copying SASMACC to remote library;
          proc catalog cat=WORK.SASMACR; 
            copy out=rmtwork.SASMACC; 
          quit;
        %end;
      %else 
        %do level = 0 %to 9;
          %if (%sysfunc(exist(WORK.SASMAC&level,CATALOG))) %then
            %do;
              %if (&_GRIDDEBUG) %then
                %put Copying SASMAC&level to remote library;
              proc catalog cat=WORK.SASMAC&level; 
                copy out=rmtwork.SASMACC; 
              quit;
            %end;
        %end;

      /*-------------------------------------------------------------------
      ** Let the user know if we failed to copy any macros
      */
      %if (%sysfunc(exist(RMTWORK.SASMACC,CATALOG)) = 0) %then
        %do;
          %put ;
          %put ;
          %put *********************************************************************;
          %put GRID: No macros were found to copy to the grid;
          %put *********************************************************************;
          %put ;
        %end;

      /*-------------------------------------------------------------------
      ** Send the client's WORK directory path to remote
      */
      %let _GRIDCLIENTWORKPATH=%sysfunc(pathname(WORK));
      %syslput _GRIDCLIENTWORKPATH=&_GRIDCLIENTWORKPATH /remote=grid;

      /*-------------------------------------------------------------------
      ** Clear the RMTWORK library since the remote will map it to
      ** something else. We will recreate it after the remote has created
      ** their LIBREF
      */
      libname rmtwork clear;

      /*-------------------------------------------------------------------
      ** Perform one-time setup on the remote SAS in the grid
      */
      rsubmit grid;

        /*-----------------------------------------------------------------
        ** Enable or disable the output of the macro text
        */
        %nrstr(%%let _GRIDSOURCEOPT=%%sysfunc(getoption(source)));

        %if (&_GRIDDEBUG) %then
          %do;
            OPTIONS SOURCE;
          %end;
        %else
          %do;
            OPTIONS NOSOURCE;
          %end;

        /*-----------------------------------------------------------------
        ** Macro to start debug output based on _GRIDDEBUG
        */
        %macro gridDebugBegin;

          %global _GRIDDEBUG;

          %global _GRIDNOTESOPT;
          %global _GRIDSOURCEOPT;
          %global _GRIDSOURCE2OPT;
          %global _GRIDMPRINTOPT;

          %let _GRIDNOTESOPT   = %sysfunc(getoption(notes));
          %let _GRIDSOURCEOPT  = %sysfunc(getoption(source));
          %let _GRIDSOURCE2OPT = %sysfunc(getoption(source2));
          %let _GRIDMPRINTOPT  = %sysfunc(getoption(mprint));

          %if (&_GRIDDEBUG) %then
            %do;
              OPTIONS SOURCE;
              OPTIONS SOURCE2;
              OPTIONS NOTES;
              OPTIONS MPRINT;
            %end;
          %else
            %do;
              OPTIONS NOSOURCE;
              OPTIONS NOSOURCE2;
              OPTIONS NONOTES;
              OPTIONS NOMPRINT;
            %end;

        %mend;

        /*-----------------------------------------------------------------
        ** Macro to end debug output started by gridDebugBegin
        */
        %macro gridDebugEnd;

          %global _GRIDNOTESOPT;
          %global _GRIDSOURCEOPT;
          %global _GRIDSOURCE2OPT;
          %global _GRIDMPRINTOPT;

          options &_GRIDMPRINTOPT;
          options &_GRIDNOTESOPT;
          options &_GRIDSOURCE2OPT;
          options &_GRIDSOURCEOPT;

        %mend;

        /*-----------------------------------------------------------------
        ** Copy the macros instantiated in the client. After the 
        ** compiliation of the setupremote macro below, all client macros 
        ** should be available.
        */
        %macro copyMacs;
            %if (%sysfunc(exist(WORK.SASMACC,CATALOG))) %then
              %do;
                %if (&_GRIDDEBUG) %then
                  %put Copying SASMACC to local macro library;

                proc catalog cat=WORK.SASMACC force; 
                  copy out=WORK.SASMACR; 
                quit;
              %end;
        %mend;

        %copyMacs;

        /*-----------------------------------------------------------------
        ** One time setup base on whether GRIDWORK has been referenced
        */
        %macro setupremote;
            %global _GRIDSHARELIBREF;
            %global _GRIDSHAREPATH;
            %global _GRIDRMTWORKPATH;

            %if (%sysfunc(libref(GRIDWORK))) %then %do;

              %let _GRIDSHARELIBREF=GRIDWORK;

              /*-----------------------------------------------------------
              ** This LIST command will cause the GRIDWORK libref to be
              ** created from the environment variable created by the
              ** sasgrid script.
              */
              LIBNAME &_GRIDSHARELIBREF LIST;

              %let _GRIDSHAREPATH=%sysfunc(pathname(GRIDWORK));

              %if (&_GRIDDEBUG) %then
                %put _GRIDSHAREPATH=&_GRIDSHAREPATH; 

            %end;

            /*-------------------------------------------------------------
            ** Make sure the RMTWORK library is available on the grid as it
            ** is on the workspace server
            */
            libname rmtwork (work);
            %let _GRIDRMTWORKPATH=%sysfunc(pathname(WORK));

            /*-------------------------------------------------------------
            ** EG submits some funny var names. This makes sure they don't
            ** cause problems.
            */
            options validvarname=any;

        %mend setupremote;

        %setupremote;

        /*-----------------------------------------------------------------
        ** Macro on remote to create unique file names for ODS output
        */
        %macro createODSFilePath(FileType=, LibName=);
            %global _GRIDODSFILEPATH;
            data _null_;
                /* Gather system information */
                gridWorkPath=pathname("&LibName");
                os=symget('sysscp');
                if os = 'WIN' then sep="\"; else sep="/";

                /* Create file name and file reference */
                filePath=cats(gridWorkPath,sep,"EGODS&FileType..&FileType");
                call symput('_GRIDODSFILEPATH', cats(gridWorkPath,sep,"EGODS&FileType..&FileType"));
            run;

            %if (&_GRIDDEBUG) %then
              %put _GRIDODSFILEPATH=&_GRIDODSFILEPATH; 

        %mend;


        /*-----------------------------------------------------------------
        ** Macro to do all the required setup on the remote system
        */
        %macro gridRemotePrecode;

          %global _RESULT;
          %global _GRIDJOBNAME;

          %gridDebugBegin;

          %if (&_GRIDDEBUG) %then
            %do;
              %put *********************************************************************;
              %put GRID: Start of global macro variables on remote before grid processing:;
              %put _GLOBAL_;
              %put GRID: End of global macro variables on remote before grid processing.;
              %put *********************************************************************;
            %end;

          %if ("&_RESULT" ne "") %then
            %gridRemoteStpPrecode;
          %else
            %gridRemoteEgAmoPrecode;

          %put *********************************************************************;
          %put GRID: This code is running on %sysfunc(grdsvc_getname(""));
          %if ("&_RESULT" eq "") %then 
            %put GRID: Output destinations enabled: &_GRIDOUTPUTDESTINATIONS;
          %put GRID: WORK is &_GRIDCLIENTWORKPATH;
          %put GRID: RMTWORK is &_GRIDRMTWORKPATH;
          %put GRID: GRIDWORK is &_GRIDSHAREPATH;
          %put GRID: Grid job is named "&_GRIDJOBNAME";
          %put *********************************************************************;
          %put ;

          %gridDebugEnd;
        %mend;

        /*-----------------------------------------------------------------
        ** Macro to do all the required setup on the remote system
        **   for Enterprise Guide/AMO
        */
        %macro gridRemoteEgAmoPrecode;

          %global _ENCODING_HTML;
          %global _ENCODING_RTF;
          %global _ENCODING_SRX;
          %global _GOPT_DEVICE;
          %global _GOPT_GFOOTNOTE;
          %global _GOPT_GTITLE;
          %global _GOPT_XPIXELS;
          %global _GOPT_YPIXELS;
          %global _ODSDEST_HTML;
          %global _ODSDEST_LISTING;
          %global _ODSDEST_PDF;
          %global _ODSDEST_RTF;
          %global _ODSDEST_SRX;
          %global _ODSOPTIONS_GRAPHCODEBASE;
          %global _ODSSTYLE_HTML;
          %global _ODSSTYLE_PDF;
          %global _ODSSTYLE_RTF;
          %global _ODSSTYLE_SRX;
          %global _ODSSTYLESHEET_HTML;
          %global _ODSSTYLESHEET_SRX;

          %local  _GRIDODSOPTIONS;

          %global _GRIDOUTPUTDESTINATIONS;

          %if (&_GRIDDEBUG) %then
            %put In gridRemoteEgAmoPrecode;

          ODS _ALL_ CLOSE;

          /*---------------------------------------------------------------
          ** ODS does not handle macros properly so we must unquote them
          */
          %let _ENCODING_HTML            =%STR(%%)UNQUOTE(&_ENCODING_HTML);           
          %let _ENCODING_RTF             =%STR(%%)UNQUOTE(&_ENCODING_RTF);             
          %let _ENCODING_SRX             =%STR(%%)UNQUOTE(&_ENCODING_SRX);             
          %let _GOPT_DEVICE              =%STR(%%)UNQUOTE(&_GOPT_DEVICE);              
          %let _GOPT_GFOOTNOTE           =%STR(%%)UNQUOTE(&_GOPT_GFOOTNOTE);           
          %let _GOPT_GTITLE              =%STR(%%)UNQUOTE(&_GOPT_GTITLE);              
          %let _GOPT_XPIXELS             =%STR(%%)UNQUOTE(&_GOPT_XPIXELS);             
          %let _GOPT_YPIXELS             =%STR(%%)UNQUOTE(&_GOPT_YPIXELS);             
          %let _ODSDEST_HTML             =%STR(%%)UNQUOTE(&_ODSDEST_HTML);             
          %let _ODSDEST_LISTING          =%STR(%%)UNQUOTE(&_ODSDEST_LISTING);          
          %let _ODSDEST_PDF              =%STR(%%)UNQUOTE(&_ODSDEST_PDF);              
          %let _ODSDEST_RTF              =%STR(%%)UNQUOTE(&_ODSDEST_RTF);              
          %let _ODSDEST_SRX              =%STR(%%)UNQUOTE(&_ODSDEST_SRX);              
          %let _ODSOPTIONS_GRAPHCODEBASE =%STR(%%)UNQUOTE(&_ODSOPTIONS_GRAPHCODEBASE); 
          %let _ODSSTYLE_HTML            =%STR(%%)UNQUOTE(&_ODSSTYLE_HTML);            
          %let _ODSSTYLE_PDF             =%STR(%%)UNQUOTE(&_ODSSTYLE_PDF);             
          %let _ODSSTYLE_RTF             =%STR(%%)UNQUOTE(&_ODSSTYLE_RTF);             
          %let _ODSSTYLE_SRX             =%STR(%%)UNQUOTE(&_ODSSTYLE_SRX);             
          %let _ODSSTYLESHEET_HTML       =%STR(%%)UNQUOTE(&_ODSSTYLESHEET_HTML);       
          %let _ODSSTYLESHEET_SRX        =%STR(%%)UNQUOTE(&_ODSSTYLESHEET_SRX);        
          
          /*---------------------------------------------------------------
          ** Initialize the list of output destinations to nothing
          */
          %let _GRIDOUTPUTDESTINATIONS=;

          /*---------------------------------------------------------------
          ** Setup graphics output
          */
          %if ("&_GOPT_DEVICE" ne "") %then
            %do;
              OPTIONS DEV=&_GOPT_DEVICE;
              GOPTIONS XPIXELS=&_GOPT_XPIXELS YPIXELS=&_GOPT_YPIXELS;
            %end;

          /*---------------------------------------------------------------
          ** Setup text (listing) output
          */
          %if ("&_ODSDEST_LISTING" ne "") %then
            %do;
              %let _GRIDOUTPUTDESTINATIONS=&_GRIDOUTPUTDESTINATIONS 'LISTING';
              %createODSFilePath(FileType=TXT, LibName=&_GRIDSHARELIBREF);
              FILENAME EGTEXT "&_GRIDODSFILEPATH";

              ODS LISTING FILE=EGTEXT GPATH="&_GRIDSHAREPATH";
            %end;

          /*---------------------------------------------------------------
          ** Setup HTML output
          */
          %if ("&_ODSDEST_HTML" ne "") %then
            %do;
              %let _GRIDOUTPUTDESTINATIONS=&_GRIDOUTPUTDESTINATIONS 'HTML';
              %createODSFilePath(FileType=HTML, LibName=&_GRIDSHARELIBREF);
              FILENAME EGHTML "&_GRIDODSFILEPATH";

              %let _GRIDODSOPTIONS=;
              %if (%sysfunc(symexist(_ODSSTYLESHEET_HTML))) %then %let _GRIDODSOPTIONS=&_GRIDODSOPTIONS STYLESHEET=&_ODSSTYLESHEET_HTML;
              %if (%sysfunc(symexist(_ODSSTYLE_HTML)))      %then %let _GRIDODSOPTIONS=&_GRIDODSOPTIONS STYLE=&_ODSSTYLE_HTML;
              %if (%sysfunc(symexist(_ENCODING_HTML)))      %then %let _GRIDODSOPTIONS=&_GRIDODSOPTIONS ENCODING="&_ENCODING_HTML";

              ODS &_ODSDEST_HTML(ID=EGHTML) FILE=EGHTML &_ODSOPTIONS_GRAPHCODEBASE GPATH="&_GRIDSHAREPATH" &_GRIDODSOPTIONS;
            %end;

          /*---------------------------------------------------------------
          ** Setup PDF output
          */
          %if ("&_ODSDEST_PDF" ne "") %then
            %do;
              %let _GRIDOUTPUTDESTINATIONS=&_GRIDOUTPUTDESTINATIONS 'PDF';
              %createODSFilePath(FileType=PDF, LibName=&_GRIDSHARELIBREF);
              FILENAME EGPDF "&_GRIDODSFILEPATH";

              %let _GRIDODSOPTIONS=;
              %if (%sysfunc(symexist(_ODSSTYLE_PDF))) %then %let _GRIDODSOPTIONS=&_GRIDODSOPTIONS STYLE=&_ODSSTYLE_PDF;

              ODS &_ODSDEST_PDF(ID=EGPDF) FILE=EGPDF &_GRIDODSOPTIONS SAS;
            %end;

          /*---------------------------------------------------------------
          ** Setup RTF output
          */
          %if ("&_ODSDEST_RTF" ne "") %then
            %do;
              %let _GRIDOUTPUTDESTINATIONS=&_GRIDOUTPUTDESTINATIONS 'RTF';
              %createODSFilePath(FileType=RTF, LibName=&_GRIDSHARELIBREF);
              FILENAME EGRTF "&_GRIDODSFILEPATH";

              %let _GRIDODSOPTIONS=;
              %if (%sysfunc(symexist(_ODSSTYLE_RTF))) %then %let _GRIDODSOPTIONS=&_GRIDODSOPTIONS STYLE=&_ODSSTYLE_RTF;
              %if (%sysfunc(symexist(_ENCODING_RTF))) %then %let _GRIDODSOPTIONS=&_GRIDODSOPTIONS ENCODING="&_ENCODING_RTF";

              ODS &_ODSDEST_RTF(ID=EGRTF) FILE=EGRTF &_GOPT_GTITLE &_GOPT_GFOOTNOTE &_GRIDODSOPTIONS;
            %end;

          /*---------------------------------------------------------------
          ** Setup SRX (SAS Report) output
          */
          %if ("&_ODSDEST_SRX" ne "") %then
            %do;
              %let _GRIDOUTPUTDESTINATIONS=&_GRIDOUTPUTDESTINATIONS 'SAS Report';
              %createODSFilePath(FileType=SRX, LibName=&_GRIDSHARELIBREF);

              FILENAME EGSR "&_GRIDODSFILEPATH";

              %let _GRIDODSOPTIONS=;
              %if (%sysfunc(symexist(_ODSSTYLESHEET_SRX))) %then %let _GRIDODSOPTIONS=&_GRIDODSOPTIONS STYLESHEET=&_ODSSTYLESHEET_SRX;
              %if (%sysfunc(symexist(_ODSSTYLE_SRX)))      %then %let _GRIDODSOPTIONS=&_GRIDODSOPTIONS STYLE=&_ODSSTYLE_SRX;
              %if (%sysfunc(symexist(_ENCODING_SRX)))      %then %let _GRIDODSOPTIONS=&_GRIDODSOPTIONS ENCODING="&_ENCODING_SRX";

              ODS &_ODSDEST_SRX(ID=EGSR) FILE=EGSR &_GOPT_GFOOTNOTE &_GOPT_GTITLE GPATH="&_GRIDSHAREPATH" &_GRIDODSOPTIONS;
            %end;

        %mend;

        /*-----------------------------------------------------------------
        ** Macro to do all the required setup on the remote system
        **   for Stored Processes
        */
        %macro gridRemoteStpPrecode;

          %global _RESULT;
          %global _ARCHIVE_PATH;
          %global _ARCHIVE_NAME;
          %global _ARCHIVE_FULLPATH;
          %global _ODSOPTIONS;

          %global _GRID_RESULT;
          %global _GRID_ARCHIVE_PATH;
          %global _GRID_ARCHIVE_NAME;
          %global _GRID_ARCHIVE_FULLPATH;

          %if (&_GRIDDEBUG) %then
            %put In gridRemoteStpPrecode;

          %let _GRID_RESULT=&_RESULT;

          %if ("&_ARCHIVE_PATH" ne "") %then
            %let _GRID_ARCHIVE_PATH=%UNQUOTE(&_ARCHIVE_PATH);
          %else
            %let _GRID_ARCHIVE_PATH=;

          %if ("&_ARCHIVE_NAME" ne "") %then
            %let _GRID_ARCHIVE_NAME=%UNQUOTE(&_ARCHIVE_NAME);
          %else
            %let _GRID_ARCHIVE_NAME=;

          %if ("&_ARCHIVE_FULLPATH" ne "") %then
            %let _GRID_ARCHIVE_FULLPATH=%UNQUOTE(&_ARCHIVE_FULLPATH);
          %else
            %let _GRID_ARCHIVE_FULLPATH=;

          %let _RESULT=PACKAGE_TO_ARCHIVE;
          %let _ARCHIVE_PATH=%sysfunc(pathname(gridwork));
          %if ("&_ARCHIVE_NAME" ne "") %then
            %let _ARCHIVE_NAME=;  * <-- (in case customer set it in code);

          %let _ODSOPTIONS=%STR(%%)UNQUOTE(&_ODSOPTIONS);

        %mend;


        /*-----------------------------------------------------------------
        ** Macro to do all the required wrapup on the remote system
        */
        %macro gridRemotePostcode;

          %global _RESULT;

          %gridDebugBegin;

          %if ("&_RESULT" ne "") %then
            %gridRemoteStpPostcode;
          %else
            %gridRemoteEgAmoPostcode;

          %if (&_GRIDDEBUG) %then
            %do;
              %put *********************************************************************;
              %put GRID: Start of global macro variables on remote after grid processing:;
              %put _GLOBAL_;
              %put GRID: End of global macro variables on remote after grid processing.;
              %put *********************************************************************;
            %end;

          %gridDebugEnd;
        %mend;

        /*-----------------------------------------------------------------
        ** Macro to do all the required wrapup on the remote system
        **   for Enterprise Guide/AMO
        */
        %macro gridRemoteEgAmoPostcode;

          %if (&_GRIDDEBUG) %then
            %put In gridRemoteEgAmoPostcode;

          ODS _ALL_ CLOSE;

        %mend;

        /*-----------------------------------------------------------------
        ** Macro to do all the required wrapup on the remote system
        **   for Stored Processes
        */
        %macro gridRemoteStpPostcode;

          %global _GRID_ARCHIVE_NAME;
          %global _GRID_ARCHIVE_PATH;
          %global _GRID_ARCHIVE_FULLPATH;

          %local lastPos;
          %local slashPos;
          %local fileSep;

          %if (&_GRIDDEBUG) %then
            %put In gridRemoteStpPostcode;

          /*
                Client set
             fullpath path name
                N       Y    Y   GENERATE FULLPATH from CLIENTPATH, CLIENTNAME
                N       Y    N   GENERATE FULLPATH from CLIENTPATH, GRIDNAME
                N       N    Y   GENERATE FULLPATH from CLIENTWORK, CLIENTNAME 
                N       N    N   GENERATE FULLPATH from CLIENTWORK, GRIDNAME
                Y       *    *   USE FULLPATH, GENERATE PATH, GENERATE NAME
          */

          /* CLIENT did not set _ARCHIVE_FULLPATH, create it */
          %if ("&_GRID_ARCHIVE_FULLPATH" eq "") %then
            %do;
              %let _ARCHIVE_PATH_LEN    =%sysfunc(lengthc(&_ARCHIVE_PATH));
              %let _ARCHIVE_FULLPATH_LEN=%sysfunc(lengthc(&_ARCHIVE_FULLPATH));
              %let _ARCHIVE_NAME        = %substr(&_ARCHIVE_FULLPATH,%eval(&_ARCHIVE_PATH_LEN+2),%eval(&_ARCHIVE_FULLPATH_LEN-&_ARCHIVE_PATH_LEN));

              /* CLIENT did not set _ARCHIVE_PATH, use client WORK path */
              %if ("&_GRID_ARCHIVE_PATH" eq "") %then
                %let _GRID_ARCHIVE_PATH = &_GRIDCLIENTWORKPATH;

              /* CLIENT did not set _ARCHIVE_NAME, use grid archive file name */
              %if ("&_GRID_ARCHIVE_NAME" eq "") %then
                %let _GRID_ARCHIVE_NAME = &_ARCHIVE_NAME;

              %if (%sysfunc(index('\\',&_GRID_ARCHIVE_PATH)) eq 0) %then
                %let fileSep=\;
              %else
                %let fileSep=/;

              %let _GRID_ARCHIVE_FULLPATH=&_GRID_ARCHIVE_PATH.&fileSep.&_GRID_ARCHIVE_NAME;
            %end;

          /* CLIENT did set _ARCHIVE_FULLPATH, parse it */
          %else 
            %do;
              %if (%sysfunc(index('\\',&_GRID_ARCHIVE_FULLPATH)) eq 0) %then
                %let fileSep=\;
              %else
                %let fileSep=/;

              %let slashPos=%sysfunc(find(&_GRID_ARCHIVE_FULLPATH,&fileSep,1));
              %do %while(&slashPos ne 0);
                  %let lastPos=%eval(&slashPos);
                  %let slashPos=%sysfunc(find(&_GRID_ARCHIVE_FULLPATH,&fileSep,%eval(&lastPos+1)));
              %end;

              %let _ARCHIVE_PATH_LEN    =%eval(&lastPos-1);
              %let _ARCHIVE_FULLPATH_LEN=%sysfunc(lengthc(&_GRID_ARCHIVE_FULLPATH));
              %let _GRID_ARCHIVE_PATH   = %substr(&_GRID_ARCHIVE_FULLPATH,1,%eval(&_ARCHIVE_PATH_LEN));
              %let _GRID_ARCHIVE_NAME   = %substr(&_GRID_ARCHIVE_FULLPATH,%eval(&_ARCHIVE_PATH_LEN+2),%eval(&_ARCHIVE_FULLPATH_LEN-&_ARCHIVE_PATH_LEN));

            %end;

          %sysrput _RESULT=&_RESULT;
          %sysrput _STPERROR=&_STPERROR;
          %sysrput _ARCHIVE_NAME=&_GRID_ARCHIVE_NAME;
          %sysrput _ARCHIVE_PATH=&_GRID_ARCHIVE_PATH;
          %sysrput _ARCHIVE_FULLPATH=&_GRID_ARCHIVE_FULLPATH;

          proc download infile="&_ARCHIVE_FULLPATH" outfile="&_GRID_ARCHIVE_FULLPATH" binary status=no; 
          run;

        %mend;

        /*-----------------------------------------------------------------
        ** Restore the normal output for code the customer submitted
        */
        options &_GRIDSOURCEOPT;

      endrsubmit;


      /*-------------------------------------------------------------------
      ** Remap the grid server's work directory to RMTWORK so the user 
      ** can see stuff in the grid work directory. This is done a second
      ** to to override any mapping done by CONNECT events from the server
      */
      libname rmtwork slibref=work server=grid;

      /*-------------------------------------------------------------------
      ** Macro on EG client to pass macro vars to server sessions
      */
      %macro gridSetMacroVarsClient;

        proc sort data=sashelp.vmacro out=vmacro nodupkey;
          by scope name;
          where scope ne 'AUTOMATIC';
        run;

        filename tmp temp lrecl=32000;

        data _null_;
          file tmp;
          set vmacro;

          strip_name = strip(trim(name));
          sas_line = cats(strip_name,' = %NRBQUOTE(&',strip_name,');');
          put '%syslput ' sas_line;
          
        run;

        %inc tmp;

        proc datasets nolist nowarn lib=work;
            delete vmacro;
        quit;

        filename tmp clear;

      %mend;

      /*-------------------------------------------------------------------
      ** Macro on EG Client to create INHERITLIB parameter from active 
      ** librefs that are not pre-assigned.
      */
      %macro gridSetLibsClient;
        %global _GridInheritLibs;

        Proc sql noprint;
          select distinct libname into :_GridInheritLibs separated by ', '
          from dictionary.libnames
          where engine ^= 'V9' and libname ^= 'SASADMIN';
        quit;

        data _null_;
          length lib           $32
                 libs          $2000
                 uri_in        $1000
                 isPreAssigned $1;

          isPreAssigned='0';
          do i = 1 to countw("&_GridInheritLibs");
            lib    = scan("&_GridInheritLibs",i);
            uri_in = cat("omsobj:SASLibrary?@Libref='", strip(lib),"'");
            rc     = metadata_getattr(uri_in,"IsPreassigned",IsPreassigned);
            if rc=0 and strip(isPreAssigned)='0' then do;
              if lengthn(libs) > 0 then libs = strip(libs) || ' ';
                libs = strip(libs) || ' ' || strip(lib);
              end;
          end;
          call symput('_GridInheritLibs',libs);
        run;
       
        %if ("&_GridInheritLibs" ne "") %then %do;
          %let _GridInheritLibs=inheritlib=(&_GridInheritLibs);
        %end;

      %mend;

      %gridDebugEnd;

    %end;

%mend gridOneTimeSetup;

/*-------------------------------------------------------------------------
** Macro to do all the grid preprocessing
*/
%macro gridClientPrecode(server=,jobname=,debug=n);

  %global _GRIDSETUPCOMPLETED;
  %global _GRIDDEBUG;
  %local  _quotelenopt;

  %let _quotelenopt= %sysfunc(getoption(QUOTELENMAX));
  options NOQUOTELENMAX;

  %gridOneTimeSetup(server=&server,jobname=&jobname,debug=&debug);

  %if ("&_GRIDSETUPCOMPLETED" eq "1") %then
    %do; 
      %gridDebugBegin;
      %gridSetMacroVarsClient;
      %gridSetLibsClient;

      %if (&_GRIDDEBUG) %then
        %do;
          %put *********************************************************************;
          %put GRID: Start of global macro variables on client before grid processing:;
          %put _GLOBAL_;
          %put GRID: End of global macro variables on client before grid processing.;
          %put *********************************************************************;
        %end;

      %gridDebugEnd;
    %end;
  %else
    %do;
      %global _GridInheritLibs;
      %put *********************************************************************;
      %put GRID: This code is NOT running on the grid. Check the initialization;
      %put       log for possible errors.;
      %put *********************************************************************;
      %put ;
      %let _GRIDINHERITLIBS=   ;
    %end;

  options &_quotelenopt;

%mend;

/*-------------------------------------------------------------------------
** Macro to do all the required wrapup on the client system
*/
%macro gridClientPostcode(signoff=n);

  %global _GRIDSETUPCOMPLETED;
  %global _RESULT;

  %if ("&_GRIDSETUPCOMPLETED" eq "1") %then
    %do;
      %gridDebugBegin;

      %if ("&_RESULT" ne "") %then
        %gridClientStpPostcode;
      %else
        %gridClientEgAmoPostcode;

      %if ("&signoff" eq "y") or ("&signoff" eq "Y") %then
        %do;
          %put GRID: Signing off;
          signoff grid;
          %let _GRIDSETUPCOMPLETED=;
        %end;

      %if (&_GRIDDEBUG) %then
        %do;
          %put *********************************************************************;
          %put GRID: Start of global macro variables on client after grid processing:;
          %put _GLOBAL_;
          %put GRID: End of global macro variables on client after grid processing.;
          %put *********************************************************************;
        %end;

      %gridDebugEnd;
    %end;

%mend;

/*-------------------------------------------------------------------------
** Macro to do all the required wrapup on the client system
**   for Enterprise Guide/AMO
*/
%macro gridClientEgAmoPostcode;

%mend;

/*-------------------------------------------------------------------------
** Macro to do all the required wrapup on the client system
**   for Stored Processes
*/
%macro gridClientStpPostcode;

%mend;

%macro gridMacrosLoaded;
  %put *********************************************************************;
  %put EG/AMO macro code loaded!;
  %put *********************************************************************;
%mend;

*%gridMacrosLoaded;
